#!/usr/bin/env python
"""Assigns a processing order to links and nodes in a schematic network.

ArcGIS Version:  10
Author:  Tim Whiteaker
Required Arguments:
    Link Features (feature class or feature layer)
    Node Features (feature class or feature layer)
Remarks:
    The input features must be Arc Hydro compliant, i.e., they must
    include fields prescribed by Arc Hydro for schematic links and
    nodes, and these fields must already be populated.  For links, the
    fields include HydroID, FromNodeID, and ToNodeID.  For nodes, the
    fields include HydroID.

    The processing order is written to a field named SortOrder. This
    field is added to the links and nodes if it does not exist.

    The purpose of this field is to inform the schematic processor about
    the order in which features in the schematic network should be
    processed.
    
"""

import os
import ConfigParser
import sys
import traceback

import arcpy

from schematic_exceptions import FieldNotFoundError, NullFieldError
from utils import check_field, get_path, add_gp_message
from classes import Link

config = ConfigParser.RawConfigParser()
config_path = get_path()
config_filename = os.path.join(config_path, 'config.cfg')
config.read(config_filename)

ID_FIELD = config.get('Fields', 'id')
FROM_NODE_FIELD = config.get('Fields', 'from_node')
TO_NODE_FIELD = config.get('Fields', 'to_node')
SORT_FIELD = config.get('Fields', 'sort_order')



def validate_inputs(link_class, node_class):
    """Raises FieldNotFoundError if inputs do not have the required fields.

    Arguments:
        link_class -- link layer or feature class
        node_class -- node layer or feature class

    """

    # Check links
    fields = arcpy.ListFields(link_class)
    if not check_field(fields, ID_FIELD):
        raise FieldNotFoundError(link_class, ID_FIELD)
    if not check_field(fields, FROM_NODE_FIELD):
        raise FieldNotFoundError(link_class, FROM_NODE_FIELD)
    if not check_field(fields, TO_NODE_FIELD):
        raise FieldNotFoundError(link_class, TO_NODE_FIELD)

    # Check nodes    
    fields = arcpy.ListFields(node_class)
    if not check_field(fields, ID_FIELD):
        raise FieldNotFoundError(node_class, ID_FIELD)

    
def add_sort_field(link_class, node_class):
    """Adds a field to store processing order for nodes and links.

    Arguments:
        link_class -- link layer or feature class
        node_class -- node layer or feature class

    """

    arcpy.AddField_management(link_class, SORT_FIELD, 'LONG')
    arcpy.AddField_management(node_class, SORT_FIELD, 'LONG')


def build_link_dictionary(link_class):
    """Returns a dictionary of links indexed by To Node ID.

    Arguments:
        link_class -- link layer or feature class

    Remarks:
        If Null HydroIDs are found in a link, this function raises
        NullFieldError.

    """
    
    links = {}
    link_cursor = arcpy.SearchCursor(link_class)
    row = None

    try:
        for row in link_cursor:
            # Build a link object
            link_id = row.getValue(ID_FIELD)
            if link_id is None:
                raise NullFieldError(link_class, ID_FIELD)
            from_node_id = row.getValue(FROM_NODE_FIELD)
            to_node_id = row.getValue(TO_NODE_FIELD)
            link = Link(link_id, from_node_id)

            # Add the link to the dictionary
            if to_node_id in links:
                link_list = links[to_node_id]
                link_list.append(link)
            else:
                link_list = [link]
                links[to_node_id] = link_list
    finally:
        if row:
            del row
        if link_cursor:
            del link_cursor

    return links

    
def get_sinks(link_dictionary, node_class):
    """Returns a set of node IDs which do not drain to any links.

    Arguments:
        link_dictionary -- dictionary of links indexed by To Node ID
        node_class -- node layer or feature class

    Remarks:
        If Null HydroIDs are found in a node, this function raises
        NullFieldError.
        
    """

    # Build a set of node IDs from the node class
    node_ids = []
    node_cursor = arcpy.SearchCursor(node_class)
    node = None
    
    try:
        for node in node_cursor:
            node_id = node.getValue(ID_FIELD)
            if node_id is None:
                raise NullFieldError(node_class, ID_FIELD)
            node_ids.append(node_id)
            
    finally:
        if node:
            del node
        if node_cursor:
            del node_cursor

    node_set = set(node_ids)

    # Build a set of From Node IDs from the links
    from_node_ids = []

    for k, v in link_dictionary.iteritems():
        for link in v:
            from_node_ids.append(link.from_node_id)
            
    from_node_set = set(from_node_ids)

    # If the node is a From Node, remove it from the set of nodes
    sinks = node_set - from_node_set

    return sinks


def process_link(link, link_dictionary, reverse_sort_order_list):
    """Reads information about the link to update processing order list.

    Arguments:
        link -- link object
        link_dictionary -- dictionary of links indexed by To Node ID
        reverse_sort_order_list -- list defining reverse order in which
                                   schematic features should be
                                   processed

    Remarks:
        reverse_sort_order_list will be updated by this function.
        This function adds the current link, the upstream node, and
        recursively processes any upstream links to populate
        reverse_sort_order_list.

    """
    
    # Add the link
    reverse_sort_order_list.append(link.id)
    # Add upstream node
    reverse_sort_order_list.append(link.from_node_id)
    # Process upstream links
    if link.from_node_id in link_dictionary:
        links = link_dictionary[link.from_node_id]
        for upstream_link in links:
            process_link(upstream_link,
                         link_dictionary,
                         reverse_sort_order_list)


def build_sort_order(link_dictionary, sink_list):
    """Returns a list defining processing order for schematic features.

    Arguments:
        link_dictionary -- dictionary of links indexed by To Node ID
        sink_list -- list of node IDs which do not drain into any links

    Remarks:
        The returned list contains the IDs of link and node features.
        The index of IDs in the list indicates the order in which the
        features associated with those IDs should be processed by the
        schematic network processor.  For example, the feature whose ID
        matches the first item in the list should be processed first.

    """

    reverse_sort_order = [] # IDs of elements from downstream to upstream

    # Start from sinks and work upstream
    for sink_id in sink_list:
        # Add the node
        reverse_sort_order.append(sink_id)
        # Work upstream
        if sink_id in link_dictionary:
            links = link_dictionary[sink_id]
            for link in links:
                process_link(link, link_dictionary, reverse_sort_order)

    # Reverse the sort_order so that the most upstream elements are first
    reverse_sort_order.reverse()
    return reverse_sort_order

    
def write_sort_order(sort_order_list, link_class, node_class):
    """Writes the sort order to attributes of the input classes.

    Arguments:
        sort_order_list -- list defining processing order for schematic
                           features
        link_class -- link layer or feature class
        node_class -- node layer or feature class

    """

    # Set up progress bar.  Use 1 percent increments.
    count = int(str(arcpy.GetCount_management(link_class)))
    count += int(str(arcpy.GetCount_management(node_class)))
    if count == 0:
        return
    increment = int(count/100.0)
    if increment == 0:
        increment = 1
    arcpy.SetProgressor("step", "Assigning processing order...", 
                        0, count, increment)
    i = 0

    # Process links
    link_cursor = arcpy.UpdateCursor(link_class)
    link = None

    try:
        for link in link_cursor:
            if (i % increment) == 0:
                arcpy.SetProgressorPosition(i)
            i += 1

            id = link.getValue(ID_FIELD)
            try:
                order = sort_order_list.index(id) + 1
                link.setValue(SORT_FIELD, order)
                link_cursor.updateRow(link)
            except:
                pass
            
    finally:
        if link:
            del link
        if link_cursor:
            del link_cursor

    # Process nodes
    node_cursor = arcpy.UpdateCursor(node_class)
    node = None
    
    try:
        for node in node_cursor:
            if (i % increment) == 0:
                arcpy.SetProgressorPosition(i)
            i += 1

            id = node.getValue(ID_FIELD)
            try:
                order = sort_order_list.index(id) + 1
                node.setValue(SORT_FIELD, order)
                node_cursor.updateRow(node)
            except:
                pass

        arcpy.SetProgressorPosition(i)
            
    finally:
        if node:
            del node
        if node_cursor:
            del node_cursor

    
def assign_processing_order(link_class, node_class):
    """Controller to assign processing order to schematic features.

    Arguments:
        link_class -- link layer or feature class
        node_class -- node layer or feature class

    """
    
    # Validate inputs
    count = int(str(arcpy.GetCount_management(link_class)))
    if count == 0:
        utils.add_gp_message('    Warning: No features included in ' + str(link_class), 1)
    count = int(str(arcpy.GetCount_management(node_class)))
    if count == 0:
        utils.add_gp_message('    Warning: No features included in ' + str(node_class), 1)
    validate_inputs(link_class, node_class)

    # Make sure the classes have a processing order field
    add_sort_field(link_class, node_class)

    add_gp_message('    Determining processing order...', 0)
    
    # Build a dictionary of links indexed by To Node ID
    links = build_link_dictionary(link_class)
    
    # Get IDs of nodes which are sinks, i.e., nodes that do not drain to links
    sinks = get_sinks(links, node_class)

    # Determine the sort order
    sort_order = build_sort_order(links, sinks)
    
    # Assign the sort order to the schematic features
    add_gp_message('    Assigning processing order...', 0)
    write_sort_order(sort_order, link_class, node_class)

    if is_test:
        print "Sort order (ID shown):", sort_order

if __name__ == '__main__':
    is_test = False

    if is_test:
        head, tail = os.path.split(os.getcwd())
        link_class = os.path.join(head, "_development\\test_data",
                                  "simple", "testLink.shp")
        node_class = os.path.join(head, "_development\\test_data",
                                  "simple", "testNode.shp")
    else:
        link_class = arcpy.GetParameterAsText(0)
        node_class = arcpy.GetParameterAsText(1)

    try:
        assign_processing_order(link_class, node_class)
    except:
        # Send traceback info to geoprocessor to aid debugging
        if not is_test:
            tb = sys.exc_info()[2]
            tbinfo = ''.join([t for t in traceback.format_tb(tb)])
            msg = "Error encountered. Traceback provided below.\n"
            add_gp_message(msg + tbinfo, 1)
        raise
