#!/usr/bin/env python
"""CSTR processing op for ArcGIS 10 schematic processor tool."""

from decimal import Decimal

import schematic_exceptions as ex

def process_values(row, input_values):
    """Simulates decay in a continuous-flow stirred-tank reactor.

    Arguments:
        row -- input row representing schematic feature
        input_values -- list of ValueTuple objects
    Remarks:
        ValueTuple objects have these attributes:
        value -- numerical value to process
        source_class -- class containing the feature providing the value
        feature_id -- unique identifier of the feature providing
                      the value

        This function requires the following fields in the input row:
            DecayConst
            Volume
            Qf
                          
    """

    # Get parameters from input row
    k = get_decimal_value(row, 'DecayConst', 'CSTR')
    volume = get_decimal_value(row, 'Volume', 'CSTR')
    qf = get_decimal_value(row, 'Qf', 'CSTR')

    # Get sum of received values
    vals = sum(Decimal(str(value_tuple.value)) for value_tuple in input_values)

    # Perform the CSTR calculation
    result = qf * vals / (qf + (k * volume))

    return result


def get_decimal_value(row, field_name, op_name):
    """Returns a Decimal representation of the row's field value."""
    try:
        val = row.getValue(field_name)
    except:
        msg = op_name + " op requires a '" + field_name + "' field." + \
              "  Input feature does not have this field."
        raise ex.OpError(msg)
    try:
        return Decimal(str(val))
    except:
        msg = op_name + " op requires numerical values in '" + \
              field_name + "' field.  The following value could not " + \
              "be converted to decimal: " + str(val)
        raise ex.OpError(msg)
