#!/usr/bin/env python
"""First-order decay processing op for ArcGIS 10 schematic processor tool."""

from decimal import Decimal
from math import exp

import schematic_exceptions as ex

def process_values(row, input_values):
    """Simulates first-order decay along a waterbody.

    Arguments:
        row -- input row representing schematic feature
        input_values -- list of ValueTuple objects
    Remarks:
        ValueTuple objects have these attributes:
        value -- numerical value to process
        source_class -- class containing the feature providing the value
        feature_id -- unique identifier of the feature providing
                      the value

        This function only processes the first value found.
        This function requires the following fields in the input row:
            DecayConst -- first-order constant
            TravelTime -- travel time through the waterbody represented
                          by the input row
        Units of measure for DecayConst must be the inverse of units
        for TravelTime.
                          
    """

    # Validate inputs
    validate_inputs(input_values)

    # Get parameters from input row
    v = input_values[0]
    input_class = v.source_class
    input_value = Decimal(str(v.value))
    feature_id = v.feature_id

    decay_constant = get_decimal_value(row, feature_id, 'DecayConst', 'Decay')
    travel_time = get_decimal_value(row, feature_id, 'TravelTime', 'Decay')
    result = (input_value * Decimal(str(exp(-1 * decay_constant * travel_time))))

    return result


def get_decimal_value(row, row_id, field_name, op_name):
    """Returns a Decimal representation of the row's field value."""
    try:
        val = row.getValue(field_name)
    except:
        msg = op_name + " op requires a '" + field_name + "' field." + \
              "  Feature " + row_id + " does not have this field."
        raise ex.OpError(msg)
    try:
        return Decimal(str(val))
    except:
        msg = op_name + " op requires numerical values in '" + \
              field_name + "' field.  The following value could not " + \
              "be converted to decimal: " + str(val)
        raise ex.OpError(msg)

    
def validate_inputs(input_values):
    """Checks values for valid inputs.

    Arguments:
        input_values -- list of ValueTuple objects

    """

    # Only one ValueTuple present
    if len(input_values) != 1:
        msg = "Decay op can only take one value as input.  " + \
              "Values provided: " + str(len(input_values)) + ".  " + \
              "Please make sure this op is used only for PASS behavior."
        raise ex.OpError(msg)
        

