#!/usr/bin/env python
"""Demonstrates how a processing op can use related tables."""

from decimal import Decimal

import arcpy

import schematic_exceptions as ex

def process_values(row, input_values):
    """Multiplies a time series of values by two.

    Arguments:
        row -- input row representing schematic feature
        input_values -- list of ValueTuple objects
    Remarks:
        ValueTuple objects have these attributes:
        value -- numerical value to process
        source_class -- class containing the feature providing the value
        feature_id -- unique identifier of the feature providing
                      the value

        The ID of the schematic feature is used to retrieve related records
        from an Arc Hydro TimeSeries table.  The TimeSeries table has these
        fields:
            FeatureID - ID of feature related to the given time series row
            VarID - ID of record in VariableDefinition table describing the
                    time series row (units of measure, variable name, etc.)
            TsTime - Datetime of the time series row
            TsValue - Value of the time series row
            
        For this example, assume the TimeSeries table has values of some
        variable (VarID of 1).  This op will multiply those values by two to
        produce a new time series (VarID of 2).
        
        This example includes minimal error handling for the sake of brevity.
    """

    # Get parameters from input row
    v = input_values[0]
    input_class = v.source_class
    # input value may not be needed, or it may represent a property of the
    # feature useful in the time series computation.  It's up to you.
    input_value = Decimal(str(v.value)) 
    feature_id = v.feature_id

    # Get the path to the TimeSeries table in the same geodatabase as the input
    # feature.  Obviously, you should add error handling for these and other
    # steps, for example, in case the TimeSeries table does not exist.
    input_path = arcpy.Describe(input_class).catalogPath

    # Assume data are in a file geodatabase. Modify accordingly if your case is
    # different.
    index = input_path.find('.gdb\\')
    timeseries_path = input_path[0:index+5] + "TimeSeries"

    # Build a whereclause to search for input time series for this feature.
    # We want rows where TimeSeries.FeatureID matches the ID (HydroID) of our
    # schematic link or node, and where TimeSeries.VarID is 1.
    whereclause = "FeatureID = " + str(feature_id) + " AND VarID = 1"

    # Read the related values (the input time series) into a list
    in_series = []
    dates = []
    row = None
    rows = arcpy.SearchCursor(timeseries_path, whereclause)
    try:
        for row in rows:
            in_series.append(row.TsValue)
            dates.append(row.TsTime)

    except:
        if not arcpy.GetMessages() == "":
            raise ex.OpError(arcpy.GetMessages(2))        

    finally:
        # Always delete cursors and rows to avoid unwanted dataset locks
        if row:
            del row
        if rows:
            del rows
    
    # Compute the output time series
    out_series = []
    for in_val in in_series:
        out_series.append(in_val * 2) # Place your awesome model code here

    # Write the resulting time series to the TimeSeries table
    row = None
    rows = arcpy.InsertCursor(timeseries_path)
    try:    
        for i in range(len(out_series)):
            row = rows.newRow()
            row.FeatureID = feature_id
            row.VarID = 2
            row.TsTime = dates[i]
            row.TsValue = out_series[i]
            rows.insertRow(row)

    except:
        if not arcpy.GetMessages() == "":
            raise ex.OpError(arcpy.GetMessages(2))        

    finally:
        if row:
            del row
        if rows:
            del rows

    # You have to give something back to the Schematic Processor to write to
    # the input feature's total value or passed value attribute.  In this
    # example, we just return a number indicating success.  The user should
    # look in the TimeSeries table to see the real results.
    result = 1 

    return result