#!/usr/bin/env python
"""Simulates processes on features in an Arc Hydro schematic network.

Tool Name:  Process Schematic Network
Source Name:  schematic_processor.py
ArcGIS Version:  10
Author:  Tim Whiteaker
Required Arguments:
    Link Features (feature class or feature layer)
    Link Incremental Value Field (field in Link Features)
    Link Total Value Field (field in Link Features)
    Link Passed Value Field (field in Link Features)
    Node Features (feature class or feature layer)
    Node Incremental Value Field (Node in Link Features)
    Node Total Value Field (field in Node Features)
    Node Passed Value Field (field in Node Features)
    Processing Ops (names of Python modules serving as processing ops).
        The names should not include the .py extension.
    Op Source Types (integer code indicating the type of features to be
        associated with a given processing op.  Corresponds to SrcType
        in nodes and LinkType in links.
    Op Feature Types (LINK | NODE)
    Op Behavior Types (RECEIVE | PASS)
Remarks:
    This tool iterates through nodes and links in a schematic network,
    applying RECEIVE and PASS behaviors to simulate hydrologic
    processes.  The tool assumes the network is dendritic and performs
    computations according to a predefined processing order, which is
    assumed to be from upstream to downstream.  

    The processing order is defined by a field named SortOrder.  This
    field should be populated prior to running this tool.  The Assign
    Processing Order tool can be used for this purpose, although the
    user can populate this field however the user wishes.

    The input features must be Arc Hydro compliant, i.e., they must
    include fields prescribed by Arc Hydro for schematic links and
    nodes, and these fields must already be populated.  For links,
    the fields include HydroID, FromNodeID, ToNodeID, and LinkType.
    For nodes, the fields include HydroID and SrcType.

    RECEIVE behaviors use the incremental value (if any) for a given
    feature and all values received from upstream features to produce
    a total value for the feature.  The tool includes inputs for users
    to specify the incremental and total value fields in the attribute
    tables of the link and node features.  The incremental value field
    should be populated prior to running this tool.  

    PASS behaviors use the total value for a feature to compute the
    value that is passed to downstream features.  The tool includes
    inputs for users to specify the passed value field in the attribute
    tables of the link and node features.

    Total and Passed value fields will be updated by this tool.

    The RECEIVE and PASS behaviors are defined by processings ops --
    Python modules which take some values and a given node or link
    and do some processing to determine an output value.  For example,
    A processing op could use an equation to simulate decay of bacteria
    in a link representing a river.  This represents PASS behavior
    since the result is the amount of bacteria that the link passes
    to the next downstream node in the network.  In this case, the
    processing op would access the link feature to obtain additional
    attributes such as a decay constant.  The author of a given
    processing op should clearly define what fields are required to
    run a given op.

    Processing ops MUST be located in the ops folder within the
    directory where the schematic_processor.py script is installed.

    If a processing op is provided, then it will be applied to the
    combination of source type, feature type, and behavior type at the
    same index in the tool's inputs.  Thus, the number of source types,
    feature types, and behavior types MUST MATCH the number of
    processing ops provided.

    If no processing ops are provided, then a simple accumulation
    of values downstream is applied.
    
"""

import ConfigParser
import inspect
import os
import sys
import traceback

import arcpy

import schematic_exceptions as ex
import utils
from classes import ValueTuple

config = ConfigParser.RawConfigParser()
config_path = utils.get_path()
config_filename = os.path.join(config_path, 'config.cfg')
config.read(config_filename)

ID_FIELD = config.get('Fields', 'id')
FROM_NODE_FIELD = config.get('Fields', 'from_node')
TO_NODE_FIELD = config.get('Fields', 'to_node')
SORT_FIELD = config.get('Fields', 'sort_order')
LINK_TYPE_FIELD = config.get('Fields', 'link_type')
SRC_TYPE_FIELD = config.get('Fields', 'src_type')

PROC_VALS_NAME = 'process_values'

def resolve_op_names(custom_ops):
    """Prepends 'ops.' to op names

    Arguments:
        custom_ops -- name of Python modules to serve as processing ops
    Remarks:
        All processing op modules should be placed in the ops folder
        beneath the folder that this script is in.

    """

    resolved_ops = []
    for index, op_name in enumerate(custom_ops):
        op_name = 'ops.' + op_name
        resolved_ops.append(op_name)

    return resolved_ops
    

def validate_inputs(link_class, node_class,
                    custom_ops, source_types, class_types, behavior_types):
    """Checks fields and ops for valid inputs.

    Arguments:
        link_class -- link layer or feature class
        node_class -- node layer or feature class
        custom_ops -- name of Python modules to serve as processing ops
        source_types -- integer code indicating the type of features to
                        be associated with the processing op at the same
                        index.  Corresponds to SrcType in nodes and
                        LinkType in links.
        class_types -- class of feature to be associated with the
                       processing op at the same index.  Valid values
                       include (LINK | NODE).
        behavior_types -- indicates if the processing op at the same
                          index applies to RECEIVE or PASS behavior.
                          Valid values include (RECEIVE | PASS).

    """

    # Check links
    fields = arcpy.ListFields(link_class)
    if not utils.check_field(fields, ID_FIELD):
        raise ex.FieldNotFoundError(link_class, ID_FIELD)
    if not utils.check_field(fields, FROM_NODE_FIELD):
        raise ex.FieldNotFoundError(link_class, FROM_NODE_FIELD)
    if not utils.check_field(fields, TO_NODE_FIELD):
        raise ex.FieldNotFoundError(link_class, TO_NODE_FIELD)
    if not utils.check_field(fields, LINK_TYPE_FIELD):
        raise ex.FieldNotFoundError(link_class, LINK_TYPE_FIELD)
    if not utils.check_field(fields, SORT_FIELD):
        raise ex.FieldNotFoundError(link_class, SORT_FIELD)

    # Check nodes    
    fields = arcpy.ListFields(node_class)
    if not utils.check_field(fields, ID_FIELD):
        raise ex.FieldNotFoundError(node_class, ID_FIELD)
    if not utils.check_field(fields, SRC_TYPE_FIELD):
        raise ex.FieldNotFoundError(node_class, SRC_TYPE_FIELD)
    if not utils.check_field(fields, SORT_FIELD):
        raise ex.FieldNotFoundError(node_class, SORT_FIELD)

    # Inputs related to processing ops must be of the same length
    if (len(custom_ops) != len(source_types) or
        len(custom_ops) != len(source_types) or
        len(custom_ops) != len(class_types) or
        len(custom_ops) != len(behavior_types)):
        msg = 'Inputs related to processing ops must have the same number ' \
              'of entries. Please check inputs for processing ops, op ' \
              'source types, op feature types, and op behavior types.'
        raise ex.OpArgumentError(msg)

    # Python modules related to ops must exist. This also loads all ops.
    for index, op_name in enumerate(custom_ops):
        try:
            op_module = utils.get_module(op_name)
        except ImportError as e:
            msg = "Could not open module '" + str(op_name) + "'.  " + \
                  str(e) + "."
            msg2 = "  Please make sure the module exists."
            # Maybe the user included .py in the module name
            if op_name[-3:].lower() == '.py':
                msg2 = "  Tip: Try leaving off the .py extension"
            msg += msg2
            raise ex.OpArgumentError(msg)

        # Make sure each module has a function called process_values
        if not PROC_VALS_NAME in dir(op_module):
            msg = PROC_VALS_NAME + " function not found in module " + \
                  op_name + ".  This function is required in order " + \
                  "for the module to be used as a processing op.  " + \
                  "Please check the module and try again."
            raise ex.OpArgumentError(msg)
        else:
            # Check number of arguments of process_values function
            process_values = getattr(op_module, PROC_VALS_NAME)
            argspec = inspect.getargspec(process_values)
            if len(argspec.args) != 2:
                msg = PROC_VALS_NAME + " function in module " + op_name + \
                      " has incorrect number of arguments.  Arguments " + \
                      "found: " + str(len(argspec.args)) + ".  Arguments " + \
                      "expected: 2.  Please check the module and try again."
                raise ex.OpArgumentError(msg)
        

def create_topology_dictionary(link_class):
    """Returns a dictionary of ID lists indexed by ID of receiving feature.

    Arguments:
        link_class -- link layer or feature class
    Remarks:
        Each item in the dictionary is a list of feature IDs for features
        that drain to a common downstream feature. The items are indexed
        by the ID of the common downtream feature.
        
    """

    topology = {}
    link_cursor = arcpy.SearchCursor(link_class)
    row = None
    
    try:
        for row in link_cursor:
            # Read the topology from the link attributes
            link_id = row.getValue(ID_FIELD)
            from_node_id = row.getValue(FROM_NODE_FIELD)
            to_node_id = row.getValue(TO_NODE_FIELD)

            # Handle the link
            if link_id in topology:
                id_list = topology[link_id]
                # MaybeDo: if performance is slow, just append and create a set instead
                if not from_node_id in id_list:
                    id_list.append(from_node_id)
            else:
                id_list = [from_node_id]
                topology[link_id] = id_list
                
            # Handle the node that the link drains to
            if to_node_id in topology:
                id_list = topology[to_node_id]
                if not link_id in id_list:
                    id_list.append(link_id)
            else:
                id_list = [link_id]
                topology[to_node_id] = id_list
    finally:
        if row:
            del row
        if link_cursor:
            del link_cursor

    return topology

def initialize_passed_values(link_class, node_class):
    """Returns a dictionary to store passed values.

    Arguments:
        link_class -- link layer or feature class
        node_class -- node layer or feature class
    Remarks:
        The dictionary is indexed by unique feature ID.
        All values in the dictionoary are initially set to None.

    """

    vals = {}
    counter = 0

    link_cursor = arcpy.SearchCursor(link_class)
    row = None
    
    try:
        for row in link_cursor:
            link_id = row.getValue(ID_FIELD)
            if link_id is None:
                raise ex.NullFieldError(link_class, ID_FIELD)
            else:
                vals[link_id] = None
                counter += 1
    finally:
        if row:
            del row
        if link_cursor:
            del link_cursor

    node_cursor = arcpy.SearchCursor(node_class)
    
    try:
        for row in node_cursor:
            node_id = row.getValue(ID_FIELD)
            if node_id is None:
                raise ex.NullFieldError(node_class, ID_FIELD)
            else:
                vals[node_id] = None
                counter += 1
    finally:
        if row:
            del row
        if node_cursor:
            del node_cursor

    # Make sure unique IDs were indeed unique
    if counter > len(vals):
        msg = "Duplicate unique IDs found.  Please make sure IDs are " + \
              "unique across all node and link features and try again."
        raise ex.DuplicateValueError(msg)

    return vals


def create_op_dictionary(custom_ops, source_types,
                         class_types, behavior_types):
    """Creates a dictionary of functions from processing modules.

    Arguments:
        custom_ops -- name of Python modules to serve as processing ops
        source_types -- integer code indicating the type of features to
                        be associated with the processing op at the same
                        index.  Corresponds to SrcType in nodes and
                        LinkType in links.
        class_types -- class of feature to be associated with the
                       processing op at the same index.  Valid values
                       include (LINK | NODE).
        behavior_types -- indicates if the processing op at the same
                          index applies to RECEIVE or PASS behavior.
                          Valid values include (RECEIVE | PASS).
    Remarks:
        The dictionary is indexed by the concatenation of source type,
        class type, and behavior type, e.g., 2-LINK-PASS.

    """

    ops = {}

    # Add the ops that the user supplied
    for index, op_name in enumerate(custom_ops):
        op_module = utils.get_module(op_name)
        process_values = getattr(op_module, PROC_VALS_NAME)
        s = source_types[index]
        c = class_types[index].upper()
        b = behavior_types[index].upper()
        key = s + '-' + c + '-' + b
        if not key in ops:
            ops[key] = process_values
        else:
            msg = "Duplicate combination of source type, feature type, " + \
                  "and behavior type found in inputs.  Please check " + \
                  "inputs and try again."
            raise ex.OpArgumentError(msg)

    # Add a default op
    op_module = utils.get_module('ops.accumulation')
    process_values = getattr(op_module, PROC_VALS_NAME)
    key = 'default'
    ops[key] = process_values
    
    return ops


def process_feature(feature, feature_class, upstream_class,
                    inc_field, tot_field, pass_field, type_field,
                    ops, topology, passed_values, is_link):
    """Processes a link or node in a schematic network.

    Arguments:
        feature -- the link or node row object to process
        feature_class -- pointer to the class containing the feature
        upstream_class -- pointer to the class providing upstream values
        inc_field -- incremental value field 
        tot_field -- total value field 
        pass_field -- passed value field
        type_field -- SrcType field for nodes, LinkType field for links
        ops -- dictionary functions to serve as processing ops,
               indexed by source_type-class_type-behavior_type
        topology -- dictionary of IDs of features that drain to a given
                    feature
        passed_values -- dictionary of passed values indexed by
                         feature ID
        is_link -- True if feature is a link; False if feature is a node
                          
    """

    if is_link:
        feature_type = 'LINK'
    else:
        feature_type = 'NODE'

    #---------------------------------------
    # Process RECEIVE behavior
    #---------------------------------------
    values = []
    
    # Get received values
    feature_id = feature.getValue(ID_FIELD)
    if feature_id in topology:
        for upstream_id in topology[feature_id]:
            if upstream_id in passed_values:
                received_value = passed_values[upstream_id]
            else:
                # May be using selection set and upstream feature might
                # not be selected.  Assume user does not want to process
                # unselected features.  Use value of zero.
                received_value = 0
            if received_value is None:
                msg = "Upstream feature " + str(upstream_id) + \
                      " has no value to pass to feature " + \
                      str(feature_id) + ".  Please assign sort " + \
                      "order and try again."
                raise ex.SortError(msg)
            else:
                v = ValueTuple(received_value, upstream_class, upstream_id)
                values.append(v)

    # Create a tuple for the incremental value
    inc_val = feature.getValue(inc_field)
    if not inc_val is None:
        v = ValueTuple(inc_val, feature_class, feature_id)
        values.append(v)

    # Call the processing op
    source_type = feature.getValue(type_field)
    if source_type is None:
        raise ex.NullFieldError(feature_type, type_field)
    op_key = "%s-%s-RECEIVE" % (source_type, feature_type)
    if op_key in ops:
        process_values = ops[op_key]
    else:
        process_values = ops['default']
    total_value = process_values(feature, values)

    # Update the row. Use float() in case op returned Decimal.
    feature.setValue(tot_field, float(total_value))

    """Debugging message
    m = feature_type + " " + str(feature_id) + " inc " + str(inc_val) + \
        " tot " + str(total_value)
    utils.add_gp_message(m, 1)
    """

    #---------------------------------------
    # Process PASS behavior
    #---------------------------------------
    values = []
    
    # Create a tuple for the total value
    v = ValueTuple(total_value, feature_class, feature_id)
    values.append(v)

    # Call the processing op
    op_key = "%s-%s-PASS" % (source_type, feature_type)
    if op_key in ops:
        process_values = ops[op_key]
    else:
        process_values = ops['default']
    passed_value = process_values(feature, values)
    passed_values[feature_id] = passed_value

    # Update the row. Use float() in case op returned Decimal.
    feature.setValue(pass_field, float(passed_value))

    """Debugging message
    m = feature_type + " " + str(feature_id) + " pass " + str(passed_value)
    utils.add_gp_message(m, 1)
    """


def iterate_network(link_class, link_inc_field, link_tot_field,
                    link_pass_field, node_class, node_inc_field,
                    node_tot_field, node_pass_field, ops,
                    topology, passed_values):
    """Iterates through features to process a schematic network.

    Arguments:
        link_class -- link layer or feature class
        link_inc_field -- incremental value field in link_class
        link_tot_field -- total value field in link_class
        link_pass_field -- passed value field in link_class
        node_class -- node layer or feature class
        node_inc_field -- incremental value field in node_class
        node_tot_field -- total value field in node_class
        node_pass_field -- passed value field in node_class
        ops -- dictionary functions to serve as processing ops,
               indexed by source_type-class_type-behavior_type
        topology -- dictionary of IDs of features that drain to a given
                    feature
        passed_values -- placeholder dictionary for passed values
                         indexed by feature ID
                          
    """

    # Set up progress bar.  Use 1 percent increments.
    count = int(str(arcpy.GetCount_management(link_class)))
    count += int(str(arcpy.GetCount_management(node_class)))
    if count == 0:
        return
    elif count == 1:
        msg = "Processing 1 feature..."
    else:
        msg = "Processing " + str(count) + " features..."
    increment = int(count/100.0)
    if increment == 0:
        increment = 1
    arcpy.SetProgressor("step", msg, 0, count, increment)
    i = 0

    # Create link cursor and node cursor sorted by sortorder
    link_cursor = arcpy.UpdateCursor(link_class, "", "", "", SORT_FIELD + " A")
    node_cursor = arcpy.UpdateCursor(node_class, "", "", "", SORT_FIELD + " A")
    link = link_cursor.next()
    node = node_cursor.next()
    is_link = False

    try:
        while link or node:
            if (i % increment) == 0:
                arcpy.SetProgressorPosition(i)
            i += 1

            # Process whichever current feature has the lower sort order.
            if link and node:
                link_order = link.getValue(SORT_FIELD) 
                node_order = node.getValue(SORT_FIELD)
                if link_order is None:
                    raise ex.NullFieldError(link_class, SORT_FIELD)
                if node_order is None:
                    raise ex.NullFieldError(node_class, SORT_FIELD)
                is_link = (link_order < node_order)
            else:
                is_link = (link != None)
                
            if is_link:
                process_feature(link, link_class, node_class,
                                link_inc_field, link_tot_field,
                                link_pass_field, LINK_TYPE_FIELD, ops,
                                topology, passed_values, is_link)
                link_cursor.updateRow(link)
                link = link_cursor.next()
            else:
                process_feature(node, node_class, link_class,
                                node_inc_field, node_tot_field,
                                node_pass_field, SRC_TYPE_FIELD, ops,
                                topology, passed_values, is_link)
                node_cursor.updateRow(node)
                node = node_cursor.next()
        
        arcpy.SetProgressorPosition(i)

    finally:
        if link:
            del link
        if link_cursor:
            del link_cursor
        if node:
            del node
        if node_cursor:
            del node_cursor

    
def process_schematic_network(link_class, link_inc_field, link_tot_field,
                              link_pass_field, node_class, node_inc_field,
                              node_tot_field, node_pass_field, custom_ops,
                              source_types, class_types, behavior_types):
    """Controller to process features in an Arc Hydro schematic network.

    Arguments:
        link_class -- link layer or feature class
        link_inc_field -- incremental value field in link_class
        link_tot_field -- total value field in link_class
        link_pass_field -- passed value field in link_class
        node_class -- node layer or feature class
        node_inc_field -- incremental value field in node_class
        node_tot_field -- total value field in node_class
        node_pass_field -- passed value field in node_class
        custom_ops -- name of Python modules to serve as processing ops
        source_types -- integer code indicating the type of features to
                        be associated with the processing op at the same
                        index.  Corresponds to SrcType in nodes and
                        LinkType in links.
        class_types -- class of feature to be associated with the
                       processing op at the same index.  Valid values
                       include (LINK | NODE).
        behavior_types -- indicates if the processing op at the same
                          index applies to RECEIVE or PASS behavior.
                          Valid values include (RECEIVE | PASS).
                          
    """

    # Resolve op names
    custom_ops = resolve_op_names(custom_ops)
    
    # Validate inputs
    link_count = int(str(arcpy.GetCount_management(link_class)))
    node_count = int(str(arcpy.GetCount_management(node_class)))
    if (link_count == 0) and (node_count == 0):
        utils.add_gp_message('    Warning: No features to process.  Aborting...', 1)
        return
    if link_count == 0:
        utils.add_gp_message('    Warning: No features included in ' + str(link_class), 1)
    elif node_count == 0:
        utils.add_gp_message('    Warning: No features included in ' + str(node_class), 1)
    validate_inputs(link_class, node_class,
                    custom_ops, source_types, class_types, behavior_types)
    
    utils.add_gp_message('    Preparing to process network...', 0)
    
    # Create a dictionary of IDs of features that drain to a given feature
    topology = create_topology_dictionary(link_class)
        
    # Create a placeholder dictionary for passed values indexed by feature ID
    passed_values = initialize_passed_values(link_class, node_class)

    # Create a dictionary of processing functions indexed by
    # source_type-class_type-behavior_type
    ops = create_op_dictionary(custom_ops, source_types,
                               class_types, behavior_types)

    # Process the network
    utils.add_gp_message('    Processing network...', 0)
    iterate_network(link_class, link_inc_field, link_tot_field,
                    link_pass_field, node_class, node_inc_field,
                    node_tot_field, node_pass_field, ops,
                    topology, passed_values)
    utils.add_gp_message('    Network processed...', 0)

    if is_test:
        print "Passed values:", passed_values

    
if __name__ == '__main__':
    is_test = False
    
    if is_test:
        head, tail = os.path.split(os.getcwd())
        link_class = os.path.join(head, "_development\\test_data",
                                  "simple", "testLink.shp")
        node_class = os.path.join(head, "_development\\test_data",
                                  "simple", "testNode.shp")
        link_inc_field = "IncVal"
        link_tot_field = "TotVal"
        link_pass_field = "PassVal"
        node_inc_field = "IncVal"
        node_tot_field = "TotVal"
        node_pass_field = "PassVal"
        custom_ops = ["accumulation"]
        source_types = ["2"]
        class_types = ["NODE"]
        behavior_types = ["RECEIVE"]
    else:
        link_class = arcpy.GetParameterAsText(0)
        link_inc_field = arcpy.GetParameterAsText(1)
        link_tot_field = arcpy.GetParameterAsText(2)
        link_pass_field = arcpy.GetParameterAsText(3)
        node_class = arcpy.GetParameterAsText(4)
        node_inc_field = arcpy.GetParameterAsText(5)
        node_tot_field = arcpy.GetParameterAsText(6)
        node_pass_field = arcpy.GetParameterAsText(7)
        custom_ops = utils.split_multivalue(arcpy.GetParameterAsText(8))     
        source_types = utils.split_multivalue(arcpy.GetParameterAsText(9))
        class_types = utils.split_multivalue(arcpy.GetParameterAsText(10))
        behavior_types = utils.split_multivalue(arcpy.GetParameterAsText(11))

    try:
        process_schematic_network(link_class, link_inc_field, link_tot_field,
                                  link_pass_field, node_class, node_inc_field,
                                  node_tot_field, node_pass_field, custom_ops,
                                  source_types, class_types, behavior_types)
    except:
        # Send traceback info to geoprocessor to aid debugging
        if not is_test:
            tb = sys.exc_info()[2]
            tbinfo = ''.join([t for t in traceback.format_tb(tb)])
            msg = "Error encountered. Traceback provided below.\n"
            utils.add_gp_message(msg + tbinfo, 1)
        raise
