#!/usr/bin/env python
"""Contains schematic processor error handling classes and functions.

ArcGIS Version:  10
Author:  Tim Whiteaker

"""

import arcpy

class SchematicError(Exception):
    """Base class for exceptions in the schematic_processor package."""
    pass

class FieldNotFoundError(SchematicError):
    """Raised when field is not found in a list of fields.

    Attributes:
        table_name -- name of the table that should contain the field
        field_Name -- name of the field
        
    """

    def __init__(self, table_name, field_name):
        self.table_name = table_name
        self.field_name = field_name
    def __str__(self):
        return repr(self.field_name + " not found in " + self.table_name)


class OpError(SchematicError):
    """Raised when a processing op raises an error."""


class OpArgumentError(SchematicError):
    """Raised when arguments related to processing ops are not valid."""


class NullFieldError(SchematicError):
    """Raised when a required field value is null.

    Attributes:
        table_name -- name of the table containing the field
        field_Name -- name of the field 
        
    """

    def __init__(self, table_name, field_name):
        self.table_name = table_name
        self.field_name = field_name
    def __str__(self):
        return repr("Value cannot be null for " + self.field_name + \
                    " field in " + self.table_name)

    
class DuplicateValueError(SchematicError):
    """Raised when duplicate values are found."""


class SortError(SchematicError):
    """Raised when links and nodes were not assigned the correct sort order."""
